#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <unistd.h>

#include <pipethrough.h>

/* Internal data types *************************************************/

struct writer_buffer {
	/*@temp@*/ const void *data;
	size_t size;
	size_t pos;
	int fd;
	bool ok;
};

struct reader_buffer {
	/*@null@*/ /*@only@*/ void *data;
	size_t size;
	size_t maxsize;
	int fd;
	bool ok;
};

struct select_request {
	int n;
	fd_set readfds;
	fd_set writefds;
};

/* Helper functions ****************************************************/

static int closeref(int *fd)
	/*@globals internalState, fileSystem, errno; @*/
	/*@modifies internalState, fileSystem, errno, *fd; @*/
{
	int result = close(*fd);
	*fd = -1;
	return result;
}

static void void_close(int fd)
	/*@globals internalState, fileSystem; @*/
	/*@modifies internalState, fileSystem; @*/
{
	int old_errno;
	
	if (fd != -1) {
		/*@-globs@*/ /*@-mods@*/
		old_errno = errno;
		(void) close(fd);
		errno = old_errno;
		/*@=globs@*/ /*@=mods@*/
	}
}

/* reader_buffer ******************************************************/

static void reader_buffer_init(/*@out@*/ struct reader_buffer *buf, int fd)
	/*@modifies *buf; @*/
{
	buf->data = NULL;
	buf->size = 0;
	buf->maxsize = 0;
	buf->fd = fd;
	buf->ok = true;
}

/* SPlint cannot describe realloc(3) correctly. */
/*@-usereleased@*/ /*@-compdef@*/ /*@-branchstate@*/ /*@-compmempass@*/
static ssize_t reader_buffer_read(struct reader_buffer *buf)
	/*@globals internalState, errno; @*/
	/*@modifies internalState, errno, *buf; @*/
{
	ssize_t res;

	if (buf->size == buf->maxsize) {
		size_t newsize = buf->maxsize + 4096;
		/*@only@*/ void *newdata;
		if (buf->data == NULL) {
			newdata = malloc(newsize);
		} else {
			newdata = realloc(buf->data, newsize);
		}
		if (newdata == NULL) {
			return (ssize_t) -1;
		}
		buf->data = newdata;
		newdata = NULL;
		buf->maxsize = newsize;
	}
	res = read(buf->fd, &(((char *) buf->data)[buf->size]),
	           buf->maxsize - buf->size);
	if (res == (ssize_t) -1) {
		return res;
	}
	buf->size += (ssize_t) res;
	return res;
}
/*@=usereleased@*/ /*@=compdef@*/ /*@=branchstate@*/ /*@=compmempass@*/

static void reader_buffer_handle(struct reader_buffer *buf)
	/*@globals internalState, fileSystem, errno, stderr; @*/
	/*@modifies internalState, fileSystem, errno, *stderr, *buf; @*/
{
	ssize_t rd = reader_buffer_read(buf);
	if (rd == (ssize_t) -1) {
		perror("read");
	}
	if (rd <= 0) {
		if (closeref(&(buf->fd)) == -1) {
			perror("close");
		}
		buf->ok = false;
	}
}

/*@-mustfreeonly@*/
static void reader_buffer_finalize(/*@special@*/ struct reader_buffer *buf)
	/*@modifies *buf; @*/
	/*@releases buf->data; @*/
{
	if (buf->data != NULL) {
		free(buf->data);
		buf->data = NULL;
	}
	buf->size = 0;
	buf->maxsize = 0;
	buf->fd = -1;
}
/*@=mustfreeonly@*/

/* writer_buffer ******************************************************/

static void writer_buffer_init(/*@out@*/ struct writer_buffer *buf, int fd,
                               const void *data, size_t size)
	/*@globals internalState, fileSystem, errno; @*/
	/*@modifies internalState, fileSystem, errno, *buf; @*/
{
	buf->fd = fd;
	buf->data = data;
	buf->size = size;
	buf->pos = 0;
	buf->ok = (size != 0);
	if (!buf->ok) {
		(void) closeref(&(buf->fd));
	}
}

static void writer_buffer_handle(struct writer_buffer *buf)
	/*@globals internalState, fileSystem, errno, stderr; @*/
	/*@modifies internalState, fileSystem, errno, *stderr, *buf; @*/
{
	ssize_t wr = write(buf->fd, &(((const char *) buf->data)[buf->pos]),
	                   buf->size - buf->pos);
	if (wr == (ssize_t) -1) {
		perror("write");
	} else {
		buf->pos += wr;
	}
	if (wr == (ssize_t) -1 || buf->pos == buf->size) {
		if (closeref(&(buf->fd)) == -1) {
			perror("close");
		}
		buf->ok = false;
	}
}

static void writer_buffer_finalize(/*@special@*/ struct writer_buffer *buf)
	/*@modifies *buf; @*/
{
	buf->data = NULL;
	buf->size = 0;
	buf->pos = 0;
	buf->fd = -1;
}

/***********************************************************************/

static void pipethrough_child(const char *command,
                              const int *stdin_pipe,
			      const int *stdout_pipe,
			      const int *stderr_pipe)
	/*@globals internalState, fileSystem, errno, stderr; @*/
	/*@modifies internalState, fileSystem, errno, *stderr; @*/
{
	const char *shell = getenv("SHELL");
	if (shell == NULL) {
		shell = "/bin/sh";
	}

	if (close(STDIN_FILENO) == -1
	    || dup2(stdin_pipe[0], STDIN_FILENO) == -1
	    || close(stdin_pipe[0]) == -1
	    || close(stdin_pipe[1]) == -1
	    || close(STDOUT_FILENO) == -1
	    || dup2(stdout_pipe[1], STDOUT_FILENO) == -1
	    || close(stdout_pipe[0]) == -1
	    || close(stdout_pipe[1]) == -1
	    || close(STDERR_FILENO) == -1
	    || dup2(stderr_pipe[1], STDERR_FILENO) == -1
	    || close(stderr_pipe[0]) == -1
	    || close(stderr_pipe[1]) == -1
	    || execl(shell, shell, "-c", command, NULL) == -1) {
	        perror("pipethrough_child");
	        (void) fflush(stderr);
		_exit(EXIT_FAILURE);
	}
}

static void select_request_init(/*@out@*/ struct select_request *sr)
	/*@modifies *sr; @*/
{
	FD_ZERO(&(sr->readfds));
	FD_ZERO(&(sr->writefds));
	sr->n = 0;
}

static void select_request_add_reader(struct select_request *sr,
                                      const struct reader_buffer *buf)
	/*@modifies *sr; @*/
{
	if (buf->ok) {
		FD_SET(buf->fd, &(sr->readfds));
		if (buf->fd + 1 > sr->n) {
			sr->n = buf->fd + 1;
		}
	}
}

static void select_request_add_writer(struct select_request *sr,
                                      const struct writer_buffer *buf)
	/*@modifies *sr; @*/
{
	if (buf->ok) {
		FD_SET(buf->fd, &(sr->writefds));
		if (buf->fd + 1 > sr->n) {
			sr->n = buf->fd + 1;
		}
	}
}

static void select_request_handle_writer(const struct select_request *sr,
                                         struct writer_buffer *buf)
	/*@globals internalState, fileSystem, errno, stderr; @*/
	/*@modifies internalState, fileSystem, errno, *stderr, *buf; @*/
{
	if (buf->ok && FD_ISSET(buf->fd, &(sr->writefds))) {
		writer_buffer_handle(buf);
	}
}

static void select_request_handle_reader(const struct select_request *sr,
                                         struct reader_buffer *buf)
	/*@globals internalState, fileSystem, errno, stderr; @*/
	/*@modifies internalState, fileSystem, errno, *stderr, *buf; @*/
{
	if (buf->ok && FD_ISSET(buf->fd, &(sr->readfds))) {
		reader_buffer_handle(buf);
	}
}

static int pipethrough_parent(int stdin_fd,
                              int stdout_fd,
                              int stderr_fd,
			      const     struct pipe_inbuffer  *stdin_buf,
			      /*@out@*/ struct pipe_outbuffer *stdout_buf,
			      /*@out@*/ struct pipe_outbuffer *stderr_buf)
	/*@globals internalState, fileSystem, errno, stderr; @*/
	/*@modifies internalState, fileSystem, errno, *stderr, *stdout_buf, *stderr_buf; @*/
{
	struct writer_buffer stdin_wbuf;
	struct reader_buffer stdout_rbuf;
	struct reader_buffer stderr_rbuf;
	struct select_request sr;
	int select_res;

	writer_buffer_init(&stdin_wbuf, stdin_fd, stdin_buf->data,
	                   stdin_buf->size);
	reader_buffer_init(&stdout_rbuf, stdout_fd);
	reader_buffer_init(&stderr_rbuf, stderr_fd);

	for (; stdin_wbuf.ok || stdout_rbuf.ok || stderr_rbuf.ok; ) {
		select_request_init(&sr);

	      retry_select:
		select_request_add_writer(&sr, &stdin_wbuf);
		select_request_add_reader(&sr, &stdout_rbuf);
		select_request_add_reader(&sr, &stderr_rbuf);
		select_res = select(sr.n, &(sr.readfds), &(sr.writefds),
		                    NULL, NULL);
		if (select_res == -1 && errno == EINTR) {
			goto retry_select;
		}
		if (select_res == -1 && errno == ENOMEM) {
			goto cleanup_select;
		}
		if (select_res == -1) {
			perror("select");
			exit(EXIT_FAILURE);
		}
		select_request_handle_writer(&sr, &stdin_wbuf);
		select_request_handle_reader(&sr, &stdout_rbuf);
		select_request_handle_reader(&sr, &stderr_rbuf);
	}
	stdout_buf->data = stdout_rbuf.data;
	stdout_rbuf.data = NULL;
	stdout_buf->size = stdout_rbuf.size;
	stderr_buf->data = stderr_rbuf.data;
	stderr_rbuf.data = NULL;
	stderr_buf->size = stderr_rbuf.size;
	return 0;

cleanup_select:
	writer_buffer_finalize(&stdin_wbuf);
	reader_buffer_finalize(&stdout_rbuf);
	reader_buffer_finalize(&stderr_rbuf);
	return -1;
}

extern int pipethrough(const     char                   *command,
                       const     struct pipe_inbuffer  *stdin_buf,
                       /*@out@*/ struct pipe_outbuffer *stdout_buf,
                       /*@out@*/ struct pipe_outbuffer *stderr_buf,
                       /*@out@*/ int                   *status)
	/*@globals internalState, fileSystem, errno, stderr; @*/
	/*@modifies internalState, fileSystem, errno, *stderr,
	            *stdout_buf, *stderr_buf, *status; @*/
{
	int retval = -1;
	pid_t pid;
	int stdin_pipe[2] = { -1, -1 };
	int stdout_pipe[2] = { -1, -1 };
	int stderr_pipe[2] = { -1, -1 };
	int saved_errno;

	if (pipe(stdin_pipe) == -1
	    || pipe(stdout_pipe) == -1
	    || pipe(stderr_pipe) == -1) {
		goto cleanup;
	}

	if ((pid = fork()) == (pid_t) -1) {
		goto cleanup;
	}

	if (pid == 0) {
		pipethrough_child(command, stdin_pipe, stdout_pipe,
		                  stderr_pipe);
		/*@notreached@*/
	}

	if (closeref(&stdin_pipe[0]) == -1
	    || closeref(&stdout_pipe[1]) == -1
	    || closeref(&stderr_pipe[2]) == -1) {
		goto cleanup;
	}

	retval = pipethrough_parent(stdin_pipe[1], stdout_pipe[0],
	                            stderr_pipe[0], stdin_buf,
				    stdout_buf, stderr_buf);

	/*@-branchstate@*/
	saved_errno = errno;
	if (waitpid(pid, status, 0) == (pid_t) -1) {
		pipe_outbuffer_finalize(stdout_buf);
		pipe_outbuffer_finalize(stderr_buf);
		if (retval == -1) {
			errno = saved_errno;
		}
		retval = -1;
	}
	/*@=branchstate@*/

cleanup:
	void_close(stderr_pipe[0]);
	void_close(stderr_pipe[1]);
	void_close(stdout_pipe[0]);
	void_close(stdout_pipe[1]);
	void_close(stdin_pipe[0]);
	void_close(stdin_pipe[1]);
	return retval;
}

/*@-mustfreeonly@*/ /*@-nullstate@*/
extern void pipe_outbuffer_finalize(struct pipe_outbuffer *buf)
	/*@modifies *buf; @*/
	/*@releases buf->data; @*/
	/*@ensures isnull buf->data; @*/
{
	if (buf->data != NULL) {
		free(buf->data);
		buf->data = NULL;
	}
	buf->size = 0;
}
/*@=mustfreeonly@*/ /*@=nullstate@*/
